@@ -1,27 +1,105 @@
 import torch
 import torch.nn.functional as F
 import numpy as np
-from typing import List, TypeVar, Any
+from typing import List, TypeVar, Any, Optional, Union, List
+import copy
 
 FlowData = TypeVar("FlowData", torch.Tensor, List[torch.Tensor], Any)
 
+
+def level_encode(data: torch.Tensor, num_levels=32, input_range=(0.0, 1.0), eps=1e-6):
+    scale = input_range[1] - input_range[0]
+    scaled = (data - input_range[0]) / scale
+    leveled = (scaled * (1.0 - eps) * num_levels).floor() + torch.empty_like(
+        data
+    ).uniform_()
+    return leveled / num_levels * scale + input_range[0]
+
+
 class Flow:
     def __init__(self, data: FlowData, device=None):
         self.data: FlowData = data
         if device is None:
             if not isinstance(data, torch.Tensor):
-                raise ValueError("Device could not be inferred from batch automatically, please provide manually.")
+                raise ValueError(
+                    "Device could not be inferred from batch automatically, please provide manually."
+                )
             device = data.device
         # These values should be per example in the data batch.
-        self.logpz: torch.Tensor = torch.scalar_tensor(0., dtype=torch.float32, device=device)
-        self.logdet: torch.Tensor = torch.scalar_tensor(0., dtype=torch.float32, device=device)
+        self.logpz: torch.Tensor = torch.scalar_tensor(
+            0.0, dtype=torch.float32, device=device
+        )
+        self.logdet: torch.Tensor = torch.scalar_tensor(
+            0.0, dtype=torch.float32, device=device
+        )
         self.zs: [torch.Tensor] = []
 
     def get_logp(self):
-        return self.logpz - self.logdet
+        return self.logpz + self.logdet
+
+    def get_elem_bits(self, input_data_shape, num_data_levels):
+        p_log2 = self.get_logp() / np.log(2.0)
+        p_elem_log2 = p_log2 / np.prod(input_data_shape[1:]) - np.log2(num_data_levels)
+        return -p_elem_log2
+
+    def deep_copy(self):
+        flow: Flow = copy.copy(self)
+        flow.logpz = flow.logpz.detach().clone()
+        flow.logdet = flow.logdet.detach().clone()
+        flow.data = self._copy_data(flow.data)
+        flow.zs = [self._copy_data(e) for e in flow.zs]
+        return flow
+
+    @classmethod
+    def _copy_data(cls, data: FlowData):
+        if data is None:
+            return None
+        elif isinstance(data, torch.Tensor):
+            return data.detach().clone()
+        elif isinstance(data, list):
+            return [cls._copy_data(e) for e in data]
+        else:
+            raise ValueError("")
+
+    def sample_like(
+        self,
+        distributions_or_temperatures: Union[
+            float,
+            List[float],
+            torch.distributions.Distribution,
+            List[torch.distributions.Distribution],
+        ] = 0.66,
+    ):
+        sample: Flow = self.deep_copy()
+        if len(sample.zs) == 0:
+            return sample
+        if isinstance(distributions_or_temperatures, float) or isinstance(
+            distributions_or_temperatures, torch.distributions.Distribution
+        ):
+            distributions_or_temperatures = [distributions_or_temperatures] * len(
+                self.zs
+            )
+        if isinstance(distributions_or_temperatures[0], float):
+            distributions_or_temperatures = [
+                torch.distributions.Normal(0.0, scale=temperature)
+                for temperature in distributions_or_temperatures
+            ]
+        distributions: List[
+            torch.distributions.Distribution
+        ] = distributions_or_temperatures
+        if len(distributions) != len(sample.zs):
+            raise ValueError(
+                "Invalid number of distributions provided, expected %d, got %d"
+                % (len(sample.zs), len(distributions))
+            )
+        for z, distribution in zip(sample.zs, distributions):
+            z.copy_(distribution.sample(z.shape))
+        return sample
 
 
 class FlowModule(torch.nn.Module):
+    CHECK_FLOW_EXPLOSION = True
+
     def forward(self, flow: Flow) -> Flow:
         return self.encode(flow)
 
@@ -47,6 +125,32 @@ class FlowModule(torch.nn.Module):
         raise NotImplementedError()
 
 
+def get_flow_flat_output(flow: Flow, batch_size: int) -> torch.Tensor:
+    components = (flow.data if isinstance(flow.data, list) else [flow.data]) + flow.zs
+    components = [c.view(batch_size, -1) for c in components if c is not None]
+    return torch.cat(components, dim=1)
+
+
+def calculate_jacobian(flow: Flow, module: FlowModule):
+    assert not flow.zs
+    assert isinstance(flow.data, torch.Tensor)
+    in_data: torch.Tensor = flow.data
+    in_data.requires_grad = True
+    batch_size = in_data.shape[0]
+    out_flow = module.encode(flow)
+    flat_out = get_flow_flat_output(out_flow, batch_size)
+    assert flat_out.numel() == in_data.numel()
+    grads = []
+    for i in range(flat_out.shape[1]):
+        if in_data.grad is not None:
+            in_data.grad.zero_()
+        flat_out[:, i].sum().backward(retain_graph=True)
+        assert in_data.grad is not None
+        grads.append(in_data.grad.view(batch_size, -1).detach().clone().unsqueeze(2))
+    jacobian = torch.cat(grads, dim=2)
+    return jacobian
+
+
 class FlowInverse(FlowModule):
     def __init__(self, flow_module: FlowModule):
         super().__init__()
@@ -67,27 +171,18 @@ class FlowSqueeze2D(FlowModule):
         out_h = h // 2
         out_w = w // 2
         fm_view = x.contiguous().view(b, c, out_h, 2, out_w, 2)
-        return fm_view.permute(0, 1, 3, 5, 2, 4).contiguous().view(b, 4 * c, out_h, out_w), 0.
+        return (
+            fm_view.permute(0, 1, 3, 5, 2, 4).contiguous().view(b, 4 * c, out_h, out_w),
+            0.0,
+        )
 
     def decode_(self, x):
-        return F.pixel_shuffle(x, 2), 0.
+        return F.pixel_shuffle(x, 2), 0.0
 
 
-class FlowConv1x1(FlowModule):
-    """https://arxiv.org/pdf/1807.03039.pdf"""
-
-    def __init__(self, channels):
+class _FlowConv1x1(FlowModule):
+    def __init__(self):
         super().__init__()
-        # Fixed permutation
-        permutation = torch.Tensor(channels, channels)
-        torch.nn.init.orthogonal_(permutation, gain=1)
-        A_LU, pivots = permutation.lu()
-        P, A_L, A_U = torch.lu_unpack(A_LU, pivots)
-        self.lower = torch.nn.Parameter(A_L)
-        self.upper = torch.nn.Parameter(A_U)
-        self.register_buffer('fixed_permutation', P)
-        # Permutation must have abs(determinant) equal to 1.
-        assert torch.slogdet(self._buffers['fixed_permutation']).logabsdet == 0.0
 
     def _get_inverse_2d_weight(self):
         # This is fast enough - only needed for decoding. In the final model this can be changed to a constant,
@@ -95,50 +190,123 @@ class FlowConv1x1(FlowModule):
         return torch.inverse(self._get_2d_weight())
 
     def _get_2d_weight(self):
-        return torch.matmul(torch.matmul(self._buffers['fixed_permutation'], torch.tril(self.lower)),
-                            torch.triu(self.upper))
+        raise NotImplementedError()
 
-    def _get_forward_logdet(self, x):
-        element_logdet = torch.sum(torch.log(torch.abs(torch.diag(self.lower))))
-        element_logdet += torch.sum(torch.log(torch.abs(torch.diag(self.upper))))
+    def _get_forward_logdet(self):
+        return torch.slogdet(self._get_2d_weight()).logabsdet
+
+    def __get_forward_logdet(self, x):
         # Multiply by the number of elements on which the 1x1 conv is performed.
-        for dim_size in x.shape[2:]:
-            element_logdet *= dim_size
-        return element_logdet
+        return self._get_forward_logdet() * np.prod(x.shape[2:])
 
     def encode_(self, x):
         dims = len(x.shape)
         if dims == 4:
-            return F.conv2d(x, weight=self._get_2d_weight().unsqueeze(2).unsqueeze(2)), self._get_forward_logdet(x)
+            return (
+                F.conv2d(x, weight=self._get_2d_weight().unsqueeze(2).unsqueeze(2)),
+                self.__get_forward_logdet(x),
+            )
         elif dims == 3:
-            return F.conv1d(x, weight=self._get_2d_weight().unsqueeze(2)), self._get_forward_logdet(x)
+            return (
+                F.conv1d(x, weight=self._get_2d_weight().unsqueeze(2)),
+                self.__get_forward_logdet(x),
+            )
         else:
             raise ValueError()
 
     def decode_(self, x):
         dims = len(x.shape)
         if dims == 4:
-            return F.conv2d(x,
-                            weight=self._get_inverse_2d_weight().unsqueeze(2).unsqueeze(2)), -self._get_forward_logdet(
-                x)
+            return (
+                F.conv2d(
+                    x, weight=self._get_inverse_2d_weight().unsqueeze(2).unsqueeze(2)
+                ),
+                -self.__get_forward_logdet(x),
+            )
         elif dims == 3:
-            return F.conv1d(x, weight=self._get_inverse_2d_weight().unsqueeze(2)), -self._get_forward_logdet(x)
+            return (
+                F.conv1d(x, weight=self._get_inverse_2d_weight().unsqueeze(2)),
+                -self.__get_forward_logdet(x),
+            )
         else:
             raise ValueError()
 
 
+class FlowConv1x1(_FlowConv1x1):
+    def __init__(self, channels, orthogonal_init=True):
+        super().__init__()
+        permutation = torch.Tensor(channels, channels)
+        if orthogonal_init:
+            torch.nn.init.orthogonal_(permutation, gain=1)
+        else:
+            torch.nn.init.xavier_normal_(permutation, gain=1)
+        self.weight = torch.nn.Parameter(permutation)
+
+    def _get_2d_weight(self):
+        return self.weight
+
+
+class FlowConv1x1Fixed(_FlowConv1x1):
+    def __init__(self, channels, orthogonal_init=True):
+        super().__init__()
+        permutation = torch.Tensor(channels, channels)
+        if orthogonal_init:
+            torch.nn.init.orthogonal_(permutation, gain=1)
+        else:
+            torch.nn.init.xavier_normal_(permutation, gain=1)
+        self.register_buffer(
+            "fixed_permutation_logdet", torch.slogdet(permutation).logabsdet
+        )
+        self.register_buffer("fixed_permutation_inverse", torch.inverse(permutation))
+        self.register_buffer("fixed_permutation", permutation)
+
+    def _get_2d_weight(self):
+        return self._buffers["fixed_permutation"]
+
+    def _get_forward_logdet(self):
+        return self._buffers["fixed_permutation_logdet"]
+
+    def _get_inverse_2d_weight(self):
+        return self._buffers["fixed_permutation_inverse"]
+
+
+class FlowConv1x1LU(_FlowConv1x1):
+    """https://arxiv.org/pdf/1807.03039.pdf"""
+
+    def __init__(self, channels, orthogonal_init=True):
+        super().__init__()
+        # Fixed permutation
+        permutation = torch.Tensor(channels, channels)
+        if orthogonal_init:
+            torch.nn.init.orthogonal_(permutation, gain=1)
+        else:
+            torch.nn.init.xavier_normal_(permutation, gain=1)
+        A_LU, pivots = permutation.lu()
+        P, A_L, A_U = torch.lu_unpack(A_LU, pivots)
+        self.lower = torch.nn.Parameter(A_L)
+        self.upper = torch.nn.Parameter(A_U)
+        self.register_buffer("fixed_permutation", P)
+        # Permutation must have abs(determinant) equal to 1.
+        assert torch.slogdet(self._buffers["fixed_permutation"]).logabsdet == 0.0
+
+    def _get_2d_weight(self):
+        return torch.matmul(
+            torch.matmul(self._buffers["fixed_permutation"], torch.tril(self.lower)),
+            torch.triu(self.upper),
+        )
+
+    def _get_forward_logdet(self):
+        element_logdet = torch.sum(torch.log(torch.abs(torch.diag(self.lower))))
+        element_logdet += torch.sum(torch.log(torch.abs(torch.diag(self.upper))))
+        return element_logdet
+
+
 def mean_for(x, dim):
-    for d in reversed(range(len(x.shape))):
-        if d != dim:
-            x = x.mean(d)
-    return x
+    return x.mean([e for e in range(len(x.shape)) if e != dim])
 
 
 def sum_for(x, dim):
-    for d in reversed(range(len(x.shape))):
-        if d != dim:
-            x = x.sum(d)
-    return x
+    return x.sum([e for e in range(len(x.shape)) if e != dim])
 
 
 def adapt_for(x, dim, dims):
@@ -156,7 +324,7 @@ class FlowActnorm(FlowModule):
         shape = [channels]
         self.log_stds = torch.nn.Parameter(torch.Tensor(*shape))
         self.means = torch.nn.Parameter(torch.Tensor(*shape))
-        self.register_buffer('is_initialized', torch.tensor(False))
+        self.register_buffer("is_initialized", torch.tensor(False))
 
     def _do_init(self, x):
         means = mean_for(x, dim=1)
@@ -167,7 +335,9 @@ class FlowActnorm(FlowModule):
             torch.nn.init.constant_(self.means, 0.0)
             self.log_stds.view_as(stds).copy_(torch.log(stds))
             self.means.view_as(means).copy_(means)
-            self._buffers['is_initialized'].copy_(torch.ones_like(self._buffers['is_initialized']))
+            self._buffers["is_initialized"].copy_(
+                torch.ones_like(self._buffers["is_initialized"])
+            )
 
     def _get_forward_logdet(self, x):
         logdet = torch.sum(self.log_stds)
@@ -176,21 +346,94 @@ class FlowActnorm(FlowModule):
         return -logdet
 
     def encode_(self, x):
-        if not self._buffers['is_initialized'].item():
+        if not self._buffers["is_initialized"].item():
             self._do_init(x)
         dims = len(x.shape)
         means = adapt_for(self.means, dim=1, dims=dims)
-        log_stds = adapt_for(self.log_stds, dim=1, dims=dims)
+        log_stds = soft_abs_limit(adapt_for(self.log_stds, dim=1, dims=dims), limit=3.0)
         return (x - means) * torch.exp(-log_stds), self._get_forward_logdet(x)
 
     def decode_(self, x):
         dims = len(x.shape)
         means = adapt_for(self.means, dim=1, dims=dims)
-        log_stds = adapt_for(self.log_stds, dim=1, dims=dims)
+        log_stds = soft_abs_limit(adapt_for(self.log_stds, dim=1, dims=dims), limit=3.0)
         x = x * torch.exp(log_stds) + means
         return x, -self._get_forward_logdet(x)
 
 
+class FlowActivenorm(FlowModule):
+    def __init__(self, channels, momentum=0.8):
+        super().__init__()
+        shape = [channels]
+        self.desired_log_scales = torch.nn.Parameter(torch.Tensor(*shape).zero_())
+        self.desired_mean = torch.nn.Parameter(torch.Tensor(*shape).zero_())
+        self.register_buffer("means", torch.Tensor(*shape).zero_())
+        self.register_buffer("vars", torch.Tensor(*shape).zero_() + 1.0)
+        self.register_buffer("num_stat_updates", torch.tensor(0))
+        self.momentum = momentum
+
+    def _get_actual_log_scales(self):
+        return soft_abs_limit(self.desired_log_scales, limit=5.0) - 0.5 * torch.log(
+            self._buffers["vars"]
+        )
+
+    def _get_actual_mean(self):
+        # (-mean*actual_std + desired_mean)
+        return self.desired_mean
+
+    def _update_stats(self, x):
+        return
+        means = mean_for(x, dim=1).detach()
+        vars = mean_for(x * x, dim=1).detach() - means ** 2 + 1e-6
+
+        old_weight = 0.0  # (1. - torch.pow(torch.tensor(self.momentum, device=x.device), self._buffers['num_stat_updates'])) * self.momentum
+        new_weight = 1 - self.momentum
+        means = (means * new_weight + self._buffers["means"] * old_weight) / (
+            new_weight + old_weight
+        )
+        vars = (vars * new_weight + self._buffers["vars"] * old_weight) / (
+            new_weight + old_weight
+        )
+
+        with torch.no_grad():
+            self._buffers["means"].copy_(means)
+            self._buffers["vars"].copy_(vars)
+            self._buffers["num_stat_updates"].copy_(
+                self._buffers["num_stat_updates"] + 1
+            )
+
+    def _get_forward_logdet(self, x):
+        logdet = torch.sum(self._get_actual_log_scales())
+        for dim_size in x.shape[2:]:
+            logdet *= dim_size
+        return -logdet
+
+    def encode_(self, x):
+        if self.training:
+            self._update_stats(x)
+
+        # (x - mean) / std * desired_std + desired_mean =
+        # = x * actual_scale + (-mean*actual_scale + desired_mean) =
+        # = x * actual_scale + actual_mean
+        # Where actual_std = desired_std / std
+        dims = len(x.shape)
+        means = adapt_for(self._get_actual_mean(), dim=1, dims=dims)
+        scales = adapt_for(torch.exp(self._get_actual_log_scales()), dim=1, dims=dims)
+        return x * scales - means, self._get_forward_logdet(x)
+
+    def decode_(self, x):
+        dims = len(x.shape)
+        means = adapt_for(self._get_actual_mean(), dim=1, dims=dims)
+        inv_scales = adapt_for(
+            torch.exp(-self._get_actual_log_scales()), dim=1, dims=dims
+        )
+        return (x + means) * inv_scales, -self._get_forward_logdet(x)
+
+
+def soft_abs_limit(x, limit=5.0):
+    return limit * torch.tanh(x / limit)
+
+
 class FlowAffineCoupling(FlowModule):
     """https://arxiv.org/pdf/1807.03039.pdf"""
 
@@ -203,6 +446,7 @@ class FlowAffineCoupling(FlowModule):
     def _get_biases_logscales(self, left, right):
         biases_logscales = self.nonlinear_module(left)
         biases, logscales = torch.chunk(biases_logscales, 2, dim=self.channel_dim)
+        logscales = soft_abs_limit(logscales, limit=1.1)
         assert biases.shape == right.shape, (left.shape, biases.shape, right.shape)
         assert logscales.shape == right.shape
         return biases, logscales
@@ -211,13 +455,22 @@ class FlowAffineCoupling(FlowModule):
         # Left can be 1 elem larger, as long as nonlinear_module takes care of this, then we are fine.
         left, right = torch.chunk(x, 2, dim=self.channel_dim)
         biases, logscales = self._get_biases_logscales(left, right)
-        return torch.cat([left, torch.exp(logscales) * right + biases], dim=self.channel_dim), sum_for(logscales, dim=0)
+        return (
+            torch.cat(
+                [left, torch.exp(logscales) * right + biases], dim=self.channel_dim
+            ),
+            sum_for(logscales, dim=0),
+        )
 
     def decode_(self, x):
         left, right = torch.chunk(x, 2, dim=self.channel_dim)
         biases, logscales = self._get_biases_logscales(left, right)
-        return torch.cat([left, torch.exp(-logscales) * (right - biases)], dim=self.channel_dim), -sum_for(logscales,
-                                                                                                           dim=0)
+        return (
+            torch.cat(
+                [left, torch.exp(-logscales) * (right - biases)], dim=self.channel_dim
+            ),
+            -sum_for(logscales, dim=0),
+        )
 
 
 class FlowAdditiveCoupling(FlowModule):
@@ -231,17 +484,28 @@ class FlowAdditiveCoupling(FlowModule):
 
     def encode_(self, x):
         left, right = torch.chunk(x, 2, dim=self.channel_dim)
-        return torch.cat([left, right + self.nonlinear_module(left)], dim=self.channel_dim), 0.
+        return (
+            torch.cat(
+                [left, right + self.nonlinear_module(left)], dim=self.channel_dim
+            ),
+            0.0,
+        )
 
     def decode_(self, x):
         left, right = torch.chunk(x, 2, dim=self.channel_dim)
-        return torch.cat([left, right - self.nonlinear_module(left)], dim=self.channel_dim), 0.
+        return (
+            torch.cat(
+                [left, right - self.nonlinear_module(left)], dim=self.channel_dim
+            ),
+            0.0,
+        )
 
 
 class FlowSequentialModule(FlowModule):
     def __init__(self, *flow_modules):
         assert flow_modules and all(
-            [isinstance(e, FlowModule) for e in flow_modules]), "Must provide nonzero number of FlowModules"
+            [isinstance(e, FlowModule) for e in flow_modules]
+        ), "Must provide nonzero number of FlowModules"
         super().__init__()
         for idx, module in enumerate(flow_modules):
             self.add_module(str(idx), module)
@@ -249,6 +513,13 @@ class FlowSequentialModule(FlowModule):
     def encode(self, flow: Flow):
         for module in self._modules.values():
             flow = module.encode(flow)
+            if self.CHECK_FLOW_EXPLOSION:
+                logdet = flow.logdet.mean().cpu().item()
+                logpz = flow.logpz.mean().cpu().item()
+                if abs(logdet) > 1e10:
+                    raise ValueError("Logdet exploded %f, module %s" % (logdet, module))
+                if abs(logpz) > 1e10:
+                    raise ValueError("Logpz exploded %f, module %s" % (logpz, module))
         return flow
 
     def decode(self, flow: Flow):
@@ -274,7 +545,9 @@ class FlowSplit(FlowModule):
             num_take = split_fraction * dim_size
             if not num_take.is_integer():
                 raise ValueError(
-                    "dim_size * split_fraction must be an integer (got %f instead) - cannot safely split flow otherwise." % num_take)
+                    "dim_size * split_fraction must be an integer (got %f instead) - cannot safely split flow otherwise."
+                    % num_take
+                )
             num_take = int(num_take)
             if num_take == 0:
                 split_data.append(None)
@@ -304,6 +577,7 @@ class FlowSplit(FlowModule):
         flow.data = torch.cat(to_concat, dim=self.split_dim)
         return flow
 
+
 class FlowParallelStep(FlowModule):
     def __init__(self, *flow_modules):
         super().__init__()
@@ -326,7 +600,9 @@ class FlowParallelStep(FlowModule):
         if len(flow.data) != len(self.flow_modules):
             raise ValueError("")
         data_items = flow.data
-        for data_item, flow_module in zip(reversed(data_items), reversed(self.flow_modules)):
+        for data_item, flow_module in zip(
+            reversed(data_items), reversed(self.flow_modules)
+        ):
             flow.data = data_item
             flow = flow_module.decode(flow)
             result.append(flow.data)
@@ -334,6 +610,7 @@ class FlowParallelStep(FlowModule):
         flow.data = result
         return flow
 
+
 # class FlowDropIndices(FlowModule):
 #     def __init__(self, *data_indices_to_drop):
 #         super().__init__()
@@ -367,6 +644,7 @@ class FlowParallelStep(FlowModule):
 #         flow.data = result
 #         return flow
 
+
 class FlowNoopStep(FlowModule):
     def encode(self, flow: Flow) -> Flow:
         return flow
@@ -374,43 +652,44 @@ class FlowNoopStep(FlowModule):
     def decode(self, flow: Flow) -> Flow:
         return flow
 
-class FlowTerminateGaussianPrior(FlowModule):
+
+class FlowTerminate(FlowModule):
     Log2PI = float(np.log(2 * np.pi))
 
-    def __init__(self):
+    def __init__(self, distribution: Optional[torch.distributions.Distribution] = None):
         super().__init__()
+        self.distribution: torch.distributions.Distribution = distribution if distribution is not None else torch.distributions.Normal(
+            0.0, 1.0
+        )
 
     def encode(self, flow: Flow) -> Flow:
         flow.zs.append(flow.data)
-        flow.logpz = flow.logpz + sum_for(self._logpz(flow.data), dim=0)
+        flow.logpz = flow.logpz + sum_for(self.distribution.log_prob(flow.data), dim=0)
         flow.data = None
         return flow
 
-    def _logpz(self, zs):
-        return -0.5 * (self.Log2PI + zs * zs)
-
     def decode(self, flow: Flow) -> Flow:
         if flow.data is not None:
             raise ValueError("")
         flow.data = flow.zs.pop()
-        flow.logpz = flow.logpz - sum_for(self._logpz(flow.data), dim=0)
+        flow.logpz = flow.logpz - sum_for(self.distribution.log_prob(flow.data), dim=0)
         return flow
 
 
-
-class FlowSplitGaussianPrior(FlowSequentialModule):
+class FlowSplitTerminate(FlowSequentialModule):
     def __init__(self, split_fraction=0.5, split_dim=1):
         assert 0 < split_fraction <= 1, "Invalid split fraction."
         if split_fraction == 1.0:
-            super().__init__(FlowTerminateGaussianPrior())
+            super().__init__(FlowTerminate())
         else:
-            super().__init__(FlowSplit(split_fractions=(1.0 - split_fraction, split_fraction), split_dim=split_dim),
-                   FlowParallelStep(
-                       FlowNoopStep(),
-                       FlowTerminateGaussianPrior()
-                   ),
-                   FlowInverse(FlowSplit(split_fractions=(1.0, 0.0), split_dim=split_dim))
-                )
+            super().__init__(
+                FlowSplit(
+                    split_fractions=(1.0 - split_fraction, split_fraction),
+                    split_dim=split_dim,
+                ),
+                FlowParallelStep(FlowNoopStep(), FlowTerminate()),
+                FlowInverse(FlowSplit(split_fractions=(1.0, 0.0), split_dim=split_dim)),
+            )
 
 
 class FlowGlowStep(FlowSequentialModule):
@@ -419,21 +698,35 @@ class FlowGlowStep(FlowSequentialModule):
     """
 
     def __init__(self, channels, conv_hidden_dim=None, use_affine_coupling=True):
-        conv_hidden_dim = conv_hidden_dim if conv_hidden_dim is not None else min(channels * 6, 512)
+        conv_hidden_dim = (
+            conv_hidden_dim if conv_hidden_dim is not None else min(channels * 12, 512)
+        )
         nonlinear_module = torch.nn.Sequential(
-            torch.nn.Conv2d(channels // 2, out_channels=conv_hidden_dim, kernel_size=3, padding=1),
+            torch.nn.Conv2d(
+                channels // 2, out_channels=conv_hidden_dim, kernel_size=3, padding=1
+            ),
             torch.nn.ReLU(),
-            torch.nn.Conv2d(conv_hidden_dim, out_channels=conv_hidden_dim, kernel_size=1, padding=0),
+            torch.nn.Conv2d(
+                conv_hidden_dim, out_channels=conv_hidden_dim, kernel_size=1, padding=0
+            ),
             torch.nn.ReLU(),
-            torch.nn.Conv2d(conv_hidden_dim, out_channels=channels if use_affine_coupling else channels // 2,
-                            kernel_size=3, padding=1),
+            torch.nn.Conv2d(
+                conv_hidden_dim,
+                out_channels=channels if use_affine_coupling else channels // 2,
+                kernel_size=3,
+                padding=1,
+            ),
+        )
+        torch.nn.init.constant_(
+            list(nonlinear_module._modules.values())[-1].weight, 0.0
+        )
+        coupling_module = (
+            FlowAffineCoupling if use_affine_coupling else FlowAdditiveCoupling
         )
-        torch.nn.init.constant_(list(nonlinear_module._modules.values())[-1].weight, 0.)
-        coupling_module = FlowAffineCoupling if use_affine_coupling else FlowAdditiveCoupling
         super().__init__(
             FlowActnorm(channels=channels),
-            FlowConv1x1(channels=channels),
-            coupling_module(nonlinear_module=nonlinear_module)
+            FlowConv1x1LU(channels=channels),
+            coupling_module(nonlinear_module=nonlinear_module),
         )
 
 
@@ -446,8 +739,15 @@ class FlowGlowNetwork(FlowSequentialModule):
         for glow_step_repeat in glow_step_repeats:
             modules.append(FlowSqueeze2D())
             channels *= 4
-            modules.append(FlowSequentialModule(*[FlowGlowStep(channels) for _ in range(glow_step_repeat)]))
-            modules.append(FlowSplitGaussianPrior(split_fraction=0.5, split_dim=1))
+            modules.append(
+                FlowSequentialModule(
+                    *[
+                        FlowGlowStep(channels, use_affine_coupling=False)
+                        for _ in range(glow_step_repeat)
+                    ]
+                )
+            )
+            modules.append(FlowSplitTerminate(split_fraction=0.5, split_dim=1))
             channels //= 2
-        modules[-1] = FlowTerminateGaussianPrior()
+        modules[-1] = FlowTerminate()
         super().__init__(*modules)
